/******************************************************************************
 *
 *       PIC18 C18 Secure Digital and Multimedia Cards Interface
 *
 ******************************************************************************
 * FileName:        sdmmc.c
 * Dependencies:    spi.h
 *					sdmmc.h
 *					sdmmc_lld.h
 *					string.h
 * Processor:       PIC18
 * Compiler:        C18
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 *
 * The software supplied herewith by Microchip Technology Incorporated
 * (the CompanyE for its PICmicro Microcontroller is intended and
 * supplied to you, the Companys customer, for use solely and
 * exclusively on Microchip PICmicro Microcontroller products. The
 * software is owned by the Company and/or its supplier, and is
 * protected under applicable copyright laws. All rights are reserved.
 * Any use in violation of the foregoing restrictions may subject the
 * user to criminal sanctions under applicable laws, as well as to
 * civil liability for the breach of the terms and conditions of this
 * license.
 *
 * THIS SOFTWARE IS PROVIDED IN AN AS ISECONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
*****************************************************************************/

#include "FAT16.h"
#include "sdmmc.h"
#ifdef USE_PIC18
#include <spi.h>                    //SPI library functions
#endif
#include "sdmmc_lld.h"
#include "string.h"

/******************************************************************************
 * Global Variables
 *****************************************************************************/
#ifdef USE_PIC18
	#pragma udata myBuff
#endif

BYTE gblSDMMCBuf[MEDIA_SECTOR_SIZE];

#ifdef USE_PIC18
	#pragma udata
	#pragma code
#endif

#ifdef USE_PIC18
	#pragma udata FATBuff
#endif

BYTE FATBuffer[MEDIA_SECTOR_SIZE];

#ifdef USE_PIC18
	#pragma udata
	#pragma code
#endif

DISK glbDiskData;

#define Yield Nop()

/*********************************************************************
 * sdmmc_cmdtable
 * - Provides information for all the sdmmc commands that we support
 * 
 * Notes: We turn off the CRC as soon as possible, so the commands with
 *        0xFF don't need to be calculated in runtime 
 *********************************************************************/
#ifdef USE_PIC18
	const rom typMMC_CMD sdmmc_cmdtable[] =
#else
	const typMMC_CMD sdmmc_cmdtable[] =
#endif
{
	// cmd					crc		response
	{cmdGO_IDLE_STATE,		0x95,	R1, NODATA},
	{cmdSEND_OP_COND,		0xF9,	R1, NODATA},
	{cmdSEND_CSD,			0xAF,	R1, MOREDATA},
	{cmdSEND_CID,			0x1B,	R1, MOREDATA},
	{cmdSTOP_TRANSMISSION,	0xC3,	R1, NODATA},
	{cmdSEND_STATUS,		0xAF,	R2, NODATA},
	{cmdSET_BLOCKLEN,		0xFF,	R1, NODATA},
	{cmdREAD_SINGLE_BLOCK,	0xFF,	R1, MOREDATA},
	{cmdREAD_MULTI_BLOCK,	0xFF,	R1, MOREDATA},
	{cmdWRITE_SINGLE_BLOCK,	0xFF,	R1, MOREDATA},
	{cmdWRITE_MULTI_BLOCK,	0xFF,	R1, MOREDATA}, 
	{cmdTAG_SECTOR_START,	0xFF,	R1, NODATA},
	{cmdTAG_SECTOR_END,		0xFF,	R1, NODATA},
	{cmdUNTAG_SECTOR,		0xFF,	R1, NODATA},
	{cmdTAG_ERASE_GRP_START, 0xFF,	R1, NODATA},
	{cmdTAG_ERASE_GRP_END,	0xFF,	R1, NODATA},
	{cmdUNTAG_ERASE_GRP,	0xFF,	R1, NODATA},
	{cmdERASE,				0xDF,	R1b, NODATA},
	{cmdLOCK_UNLOCK,		0x89,	R1b, NODATA},  
	{cmdSD_APP_OP_COND,		0xE5,	R1, NODATA},
	{cmdAPP_CMD,			0x73,	R1, NODATA},
	{cmdREAD_OCR,			0x25,	R3, NODATA},
	{cmdCRC_ON_OFF,			0x25,	R1, NODATA}
};

// BPB FAT12
typedef struct __BPB_FAT12 {
        SWORD BS_jmpBoot;        // Jump Command
        BYTE  BS_OEMName[8];     // OEM name
        WORD BPB_BytsPerSec;     // BYTEs per sector
        BYTE  BPB_SecPerClus;     // sectors per allocation unit
        WORD BPB_RsvdSecCnt;     // number of reserved sectors after start
        BYTE  BPB_NumFATs;        // number of FATs
        WORD BPB_RootEntCnt;     // number of root directory entries
        WORD BPB_TotSec16;       // total number of sectors
        BYTE  BPB_Media;          // media descriptor
        WORD BPB_FATSz16;         // number of sectors per FAT
        WORD BPB_SecPerTrk;      // sectors per track
        WORD BPB_NumHeads;       // number of heads
        DWORD BPB_HiddSec;        // number of hidden sectors
        BYTE  BS_Reserved0;       // Nothing
        BYTE  BS_DrvNum;          // Int 13 drive number
        BYTE  BS_Reserved1;       // Nothing
        BYTE  BS_BootSig;         // 0x29
        BYTE  BS_VolID[4];        // Volume Id
        BYTE  BS_VolLab[11];       // Volume Label
        BYTE  BS_FilSysType[8];   // File system type, not used for determination   
}_BPB_FAT12;

typedef _BPB_FAT12 * BPB_FAT12;

// BPB FAT32
typedef struct __BPB_FAT32 {
        SWORD BPB_jmpBoot;        // Jump Command
        BYTE  BPB_OEMName[8];     // OEM name
        WORD BPB_BytsPerSec;     // BYTEs per sector
        BYTE  BPB_SecPerClus;     // sectors per allocation unit
        WORD BPB_RsvdSecCnt;     // number of reserved sectors after start
        BYTE  BPB_NumFATs;        // number of FATs
        WORD BPB_RootEntCnt;     // number of root directory entries
        WORD BPB_TotSec16;       // total number of sectors
        BYTE  BPB_Media;          // media descriptor
        WORD BPB_FATSz16;         // number of sectors per FAT
        WORD BPB_SecPerTrk;      // sectors per track
        WORD BPB_NumHeads;       // number of heads
        DWORD BPB_HiddSec;        // number of hidden sectors
        DWORD BPB_TotSec32;       // New 32bit total sec count
        DWORD BPB_FATSz32;        // Sectors occupied by one FAT
        WORD BPB_ExtFlags;       // Presently active FAT as defined by bits 0-3 if bit 7 is 1
        WORD BPB_FSVers;         // FAT32 filesystem version, should be 0:0
        DWORD BPB_RootClus;       // start cluster of the root directory (should be 2)
        WORD BPB_FSInfo;         // filesystem info structure sector
        WORD BPB_BkBootSec;      // backup boot sector normally 6
        BYTE  BPB_Reserved[12];   // Reserved memory
        BYTE  BS_DrvNum;          // Int 13 drive number
        BYTE  BS_Reserved1;       // Nothing
        BYTE  BS_BootSig;         // 0x29
        BYTE  BS_VolID[4];        // Volume Id
        BYTE  BS_VolLab[11];       // Volume Label
        BYTE  BS_FilSysType[8];   // File system type, not used for determination  
}_BPB_FAT32;

typedef _BPB_FAT32 * BPB_FAT32;

// BPB FAT16
typedef struct __BPB_FAT16 {
        SWORD BS_jmpBoot;        // Jump Command
        BYTE  BS_OEMName[8];     // OEM name
        WORD BPB_BytsPerSec;     // BYTEs per sector
        BYTE  BPB_SecPerClus;     // sectors per allocation unit
        WORD BPB_RsvdSecCnt;     // number of reserved sectors after start
        BYTE  BPB_NumFATs;        // number of FATs
        WORD BPB_RootEntCnt;     // number of root directory entries
        WORD BPB_TotSec16;       // total number of sectors
        BYTE  BPB_Media;          // media descriptor
        WORD BPB_FATSz16;         // number of sectors per FAT
        WORD BPB_SecPerTrk;      // sectors per track
        WORD BPB_NumHeads;       // number of heads
        DWORD BPB_HiddSec;        // number of hidden sectors
        DWORD BPB_TotSec32;       // New 32bit total sec count
        BYTE  BS_DrvNum;          // Int 13 drive number
        BYTE  BS_Reserved1;       // Nothing
        BYTE  BS_BootSig;         // 0x29
        BYTE  BS_VolID[4];        // Volume Id
        BYTE  BS_VolLab[11];       // Volume Label
        BYTE  BS_FilSysType[8];   // File system type, not used for determination     
}_BPB_FAT16;

#define BSI_BOOTJMP			0
#define BSI_OEM_NAME		3
#define BSI_BytesPerSec		11
#define BSI_SecPerClus		13
#define BSI_RsvdSecCount 	14
#define BSI_NumFATs			16
#define BSI_RootEntCnt		17
#define BSI_TotSec16 		19
#define BSI_Media 			21
#define BSI_FATSz16 		22
#define BSI_SecPerTrk 		24
#define BSI_NumHeads 		26
#define BSI_HiddSec 		28
#define BSI_TotSec32 		32
#define BSI_DrvNum 			36
#define BSI_Reserved1 		37
#define BSI_BootSig 		38
#define BSI_VolID 			39
#define BSI_VolLab 			43
#define BSI_FilSysType 		54


typedef _BPB_FAT16 * BPB_FAT16;

// PTE_FAT - Partition Table Entry

typedef struct _PTE_FAT
{
    BYTE      PTE_BootDes;            // Boot Descriptor, 0x80
    SWORD     PTE_FrstPartSect;       // First Partion Sector
    BYTE      PTE_FSDesc;             // File System Descriptor 
    SWORD     PTE_LstPartSect;        // Last Partion Sector
    DWORD     PTE_FrstSect;           // First Sector Position
    DWORD     PTE_NumSect;            // Number of Sectors in partion
}PTE_FAT;

// PT_FAT - Partition Table 
typedef struct __PT_FAT
{
    BYTE      ConsChkRtn[446];
    PTE_FAT Partition0;
    PTE_FAT Partition1;
    PTE_FAT Partition2;
    PTE_FAT Partition3;
    BYTE      Signature0;     // 0x55
    BYTE      Signature1;     // 0xAA
}_PT_FAT;

typedef _PT_FAT *  PT_FAT;

typedef union _MMCstate
{
    struct
	{
        BYTE isSDMMC : 1;     // set if it is a SDMMC 
        BYTE isWP    : 1;     // set if it is write protected 
    };
    BYTE  _byte;
}MMCSTATE;

typedef struct __BS
{
    union
    {
        _BPB_FAT32  FAT_32;	
        _BPB_FAT16  FAT_16;	
        _BPB_FAT12  FAT_12;	
    }FAT;    
    BYTE  	Reserved[MEDIA_SECTOR_SIZE-sizeof(_BPB_FAT32)-2];
    BYTE      Signature0;     // 0x55
    BYTE      Signature1;     // 0xAA
}_BS;

typedef _BS * BS;

typedef union _CARDSTATUS
{
    struct
    {
        BYTE  CS_Inserted:1;  // Card is inserted and initialized
        BYTE  CS_Reset:1;     // Card has just been reset
        BYTE  CS_SDMMC:1;     // Card is of the SDMMC type
        BYTE  CS_WP:1;        // Card is write Protected
        BYTE  CS_Failure:1;   // There was a failure initializing the card
    };
    WORD _word;   
}CARDSTATUS;

// Master Boot Record offsets
#define FO_MBR          0L  // master boot record sector LBA

#ifndef EOF
#define EOF         (-1)
#endif

#define FAT_GOOD_SIGN_0     0x55
#define FAT_GOOD_SIGN_1     0xAA
#define BOOTABLE            0x80
#define NON_BOOTABLE        0x00

/******************************************************************************
 * Prototypes
 *****************************************************************************/

BYTE DISKmount( DISK *dsk, BYTE* b, MMCSTATE *Flag);
BYTE LoadMBR(DISK *dsk);
BYTE LoadBootSector(DISK *dsk);
extern void Delayms(BYTE milliseconds);
void StopCard(DISK *dsk, CARDSTATUS *cs);
BYTE StartupCard(DISK *dsk, CARDSTATUS *CardStatus);
BYTE ReadMedia(void);
BYTE MediaInitialize(void);
MMC_RESPONSE SendMMCCmd(BYTE cmd, DWORD address);

#ifdef USE_PIC24
	BYTE ReadByte( BYTE* pBuffer, WORD index );
	WORD ReadWord( BYTE* pBuffer, WORD index );
	DWORD ReadDWord( BYTE* pBuffer, WORD index );
	
	void OpenSPI( unsigned int sync_mode, unsigned char bus_mode, unsigned char smp_phase);
	void CloseSPI( void );
	unsigned char WriteSPI( unsigned char data_out );
#endif

/******************************************************************************
 * Function:        BYTE MediaDetect(void)
 *
 * PreCondition:    MediaInit function has been executed.
 *
 * Input:           void
 *
 * Output:          TRUE   - Card detected
 *                  FALSE   - No card detected
 *
 * Side Effects:    None
 *
 * Overview:        None
 *
 * Note:            None
 *****************************************************************************/
BYTE MediaDetect()
{
	// give it a debounce here
	Delayms(10);

	return(!SD_CD);
}//end MediaDetect


/******************************************************************************
 * Function			BYTE MediaInit (void)
 * KETTER
******************************************************************************/

BYTE MediaInit ()
{
	CARDSTATUS	cs;

	// Turn off the card
	MMC_OFF;
	SD_CD_TRIS = INPUT;				//Card Detect - input
	SD_CS = 1;							//Initialize Chip Select line
	SD_CS_TRIS = OUTPUT;				//Card Select - output
	SD_WE_TRIS = INPUT;				//Write Protect - input

	return  (StartupCard (&glbDiskData, &cs) == CE_GOOD);
}
		

/******************************************************************************
 * Function:        BYTE MediaInitialize(MMCSTATE *Flag)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          mmcValid                    - Everything is fine
 *                  mmcCardInitCommFailure      - Communication has never been established with card
 *                  mmcCardNotInitFailure       - Card did not go into an initialization phase
 *                  mmcCardInitTimeout          - Card initialization has timedout
 *                  mmcCardTypeInvalid          - Card type was not able to be defined
 *                  mmcCardBadCmd               - Card did not reconized the command
 *                  mmcCardTimeout              - Card timedout during a read, write or erase sequence
 *                  mmcCardCRCError             - A CRC error has occurred during a read, data should be invalidated
 *                  mmcCardDataRejected         - Card and data sent's CRC did not match
 *
 * Side Effects:    Variable 'gblMMC_media' is updated
 *
 * Overview:        MediaInitialize initializes the media card and supporting variables.
 *
 * Note:            goto's were used for errro conditions
 *****************************************************************************/
BYTE MediaInitialize(void)
{
	WORD timeout;
	BYTE       status = TRUE;
	MMC_RESPONSE    response; 

	SD_CS = 1;					//Initialize Chip Select line

	// This function was called for a reason Turn it on
	MMC_ON;

	//Media powers up in the open-drain mode and cannot handle a clock faster
	//than 400kHz. Initialize SPI port to slower than 400kHz
	#ifdef USE_PIC18
		OpenSPI(SPI_FOSC_64, MODE_00, SMPEND);
	#elif defined USE_PIC33
		OpenSPI((unsigned int)0x003C, 3, 0);
	#elif defined USE_PIC30
		OpenSPI((unsigned int)0x003C, 3, 0);
	#else
		OpenSPI((unsigned int)0x003C, 3, 0);
	#endif
	// let the card power on and initialize
	Delayms(100);

	//Media requires 80 clock cycles to startup [8 clocks/BYTE * 10 us]
	for(timeout=0; timeout<10; timeout++)
		mSend8ClkCycles();

	SD_CS = 0;
	
	Delayms(1);

	// Send CMD0 to reset the media
	response = SendMMCCmd(GO_IDLE_STATE,0x0);

	if(response.r1._byte == MMC_BAD_RESPONSE)
	{
		status = FALSE;      // we have not got anything back from the card
		goto InitError;
	}

	// See if the device is ready
	if((response.r1._byte & 0xF7) != 0x01)                  //0x01=No Err&Busy Initializing
	{
		status = FALSE;      // we have not got anything back from the card
		goto InitError;
	}

	// According to spec cmd1 must be repeated until MMC card is fully initialized
	timeout = 0xFFF;

	do
	{
		response = SendMMCCmd(SEND_OP_COND,0x0);
		timeout--;
	}while(response.r1._byte != 0x00 && timeout != 0);

	// see if it failed
	if(timeout == 0)
	{
		status = FALSE;      // we have not got anything back from the card
		goto InitError;
	}
	else      
  		#ifdef USE_PIC18
			OpenSPI(SPI_FOSC_4, MODE_00, SMPEND);       //Increase clock speed
		#elif defined USE_PIC33
			OpenSPI(0x003E, 3, 0);
		#elif defined USE_PIC30
			OpenSPI(0x003E, 3, 0);
		#else
			OpenSPI(0x003E, 3, 0);
		#endif


	// Turn off CRC7 if we can, might be an invalid cmd on some cards (CMD59)
	response = SendMMCCmd(CRC_ON_OFF,0x0);

	// Ok Now set the block length to 512. It should be already
	SendMMCCmd(SET_BLOCKLEN,MEDIA_SECTOR_SIZE);

    // read it a couple times until we sucessfully read it, seen san disks that really do not
	// like this init routine for somereason, almost have to prime the device...
	for(timeout = 0xFF; timeout > 0 && SECTORread(0x0,gblSDMMCBuf) != TRUE; timeout--)
	{;}

	// see if we had an issue
	if(timeout == 0)
	{
		status = FALSE;
		goto InitError;
	}

	return(status);

InitError:
	SD_CS = 1;                               // deselect the devices

	// Turn it off
	MMC_OFF;

	return(status);
}//end MediaInitialize

//----------------------------------------------------------------------
// ShutdownMedia
// Simply turn off the card
//
void ShutdownMedia(void)
{
	// close the spi bus
	CloseSPI();

	// deselect the device
	SD_CS = 1;

	// Turn off the card
	MMC_OFF;
}

/******************************************************************************
 * Function:        void SendMMCCmd(BYTE cmd, DWORD address)
 *
 *
 * Input:           None
 *
 * Output:          response            - Response from the card
 *                                      - 0x00 or 0x01 Command received 
 *                                        successfully, else, command failed
 *                  -Bit 0              - In idle state if 1
 *                  -Bit 1              - Erase Reset if 1
 *                  -Bit 2              - Illgal Command if 1
 *                  -Bit 3              - Com CRC Error if 1
 *                  -Bit 4              - Erase Sequence Error if 1
 *                  -Bit 5              - Address Error if 1
 *                  -Bit 6              - Parameter Error if 1
 *                  -Bit 7              - Not used, always '0'
 *
 * Side Effects:    None
 *
 * Overview:        SendMMCCmd prepares the command packet and sends it out
 *                  over the SPI interface. Response data of type 'R1' (see
 *                  SD or MMC product manuals) is returned.
 *
 * Note:            MMC_CS is not set at the end of this function.
 *                  if the command has no data stage, call macro
 *                  mSendMediaCmd_NoData(), it reasserts MMC_CS to 1.
 *                  If the command has a data stage, MMC_CS must be
 *                  reasserted after the data transfer stage is complete.
 *                  See SECTORread and SECTORwrite for examples.
 *****************************************************************************/
MMC_RESPONSE SendMMCCmd(BYTE cmd, DWORD address)
{
	WORD timeout = 0x8;
	BYTE index;
	MMC_RESPONSE    response;
	CMD_PACKET  CmdPacket;

	SD_CS = 0;                           //Card Select

	// Copy over data
	CmdPacket.cmd        = sdmmc_cmdtable[cmd].CmdCode;
	CmdPacket.address    = address;
	CmdPacket.crc        = sdmmc_cmdtable[cmd].CRC;       // Calc CRC here

	CmdPacket.TRANSMIT_BIT = 1;             //Set Tranmission bit

	WriteSPI(CmdPacket.cmd);                //Send Command
	WriteSPI(CmdPacket.addr3);              //Most Significant Byte
	WriteSPI(CmdPacket.addr2);
	WriteSPI(CmdPacket.addr1);
	WriteSPI(CmdPacket.addr0);              //Least Significant Byte
	WriteSPI(CmdPacket.crc);                //Send CRC

	// see if  we are going to get a response
	if(sdmmc_cmdtable[cmd].responsetype == R1 || sdmmc_cmdtable[cmd].responsetype == R1b)
	{
		do
		{
			response.r1._byte = ReadMedia();
			timeout--;
		}while(response.r1._byte == MMC_FLOATING_BUS && timeout != 0);
	}
	else if(sdmmc_cmdtable[cmd].responsetype == R2)
	{
		ReadMedia();

		response.r2._byte1 = ReadMedia();
		response.r2._byte0 = ReadMedia();
	}

	if(sdmmc_cmdtable[cmd].responsetype == R1b)
	{
		response.r1._byte = 0x00;

		for(index =0; index < 0xFF && response.r1._byte == 0x00; index++)
		{
			timeout = 0xFFFF;

			do
			{
				response.r1._byte = ReadMedia();
				timeout--;
			}while(response.r1._byte == 0x00 && timeout != 0);
		}
	}

	mSend8ClkCycles();                      //Required clocking (see spec)

	// see if we are expecting data or not
	if(!(sdmmc_cmdtable[cmd].moredataexpected))
		SD_CS = 1;

	return(response);
}


/******************************************************************************
 * Function:        BYTE SECTORread(DWORD sector_addr, BYTE *buffer)
 *
 * PreCondition:    None
 *
 * Input:           sector_addr - Sector address, each sector contains 512-byte
 *                  buffer      - Buffer where data will be stored, see
 *                                'ram_acs.h' for 'block' definition.
 *                                'Block' is dependent on whether internal or
 *                                external memory is used
 *
 * Output:          See MMC_Error.
 *
 * Side Effects:    None
 *
 * Overview:        SECTORread reads 512 bytes of data from the card starting
 *                  at the sector address specified by sector_addr and stores
 *                  them in the location pointed to by 'buffer'.
 *
 * Note:            The card expects the address field in the command packet
 *                  to be byte address. Therefore the sector_addr must first
 *                  be converted to byte address. This is accomplished by
 *                  shifting the address left 9 times.
 *****************************************************************************/
BYTE SECTORread(DWORD sector_addr, BYTE* buffer)
{
	WORD index;
	WORD delay;
	MMC_RESPONSE    response;
	BYTE data_token;
	BYTE status = TRUE;
	DWORD	new_addr;

	// send the cmd
	new_addr = sector_addr << 9;
	response = SendMMCCmd(READ_SINGLE_BLOCK,new_addr);

	// Make sure the command was accepted
	if(response.r1._byte != 0x00)
	{
		response = SendMMCCmd (READ_SINGLE_BLOCK,new_addr);
		if(response.r1._byte != 0x00)
			status = FALSE;
	}
	else
	{
		index = 0x2FF;
		
		// Timing delay- at least 8 clock cycles
		delay = 0x40;
		while (delay)
			delay--;
		
		//Now, must wait for the start token of data block
		do
		{
      		data_token = ReadMedia();
			index--;	
	
			delay = 0x40;
			while (delay)
				delay--;

		}while((data_token == MMC_FLOATING_BUS) && (index != 0));

		// Hopefully that zero is the datatoken
		if((index == 0) || (data_token != DATA_START_TOKEN))
		{
			status = FALSE;
		}
		else
		{
			for(index = 0; index < MEDIA_SECTOR_SIZE; index++)		//Reads in 512-byte of data
			{
					buffer[index] = ReadMedia();
			}
			// Now ensure CRC
			mReadCRC();					//Read 2 bytes of CRC
			//status = mmcCardCRCError;
		}

		mSend8ClkCycles();				//Required clocking (see spec)
	}

	SD_CS = 1;

	return(status);
}//end SECTORread

/******************************************************************************
 * Function:        BYTE SECTORwrite(DWORD sector_addr, BYTE *buffer)
 *
 * PreCondition:    None
 *
 * Input:           sector_addr - Sector address, each sector contains 512-byte
 *                  buffer      - Buffer where data will be read, see
 *                                'ram_acs.h' for 'block' definition.
 *                                'Block' is dependent on whether internal or
 *                                external memory is used
 *
 * Output:          See MMC_Error.
 *
 * Side Effects:    None
 *
 * Overview:        SECTORwrite sends 512 bytes of data from the location
 *                  pointed to by 'buffer' to the card starting
 *                  at the sector address specified by sector_addr.
 *
 * Note:            The card expects the address field in the command packet
 *                  to be byte address. Therefore the sector_addr must first
 *                  be converted to byte address. This is accomplished by
 *                  shifting the address left 9 times.
 *****************************************************************************/
BYTE SECTORwrite(DWORD sector_addr, BYTE* buffer)
{
	WORD index;
    BYTE data_response;
	MMC_RESPONSE    response; 
    BYTE status = TRUE;

	if (sector_addr == 0)
		status = FALSE;
	else
	{
		// send the cmd
		response = SendMMCCmd(WRITE_SINGLE_BLOCK,(sector_addr << 9));

		// see if it was accepted
		if(response.r1._byte != 0x00)
			status = FALSE;
		else
		{
			WriteSPI(DATA_START_TOKEN);                 //Send data start token

			for(index = 0; index < 512; index++)                    //Send 512 bytes of data
				WriteSPI(buffer[index]);

			// calc crc
			mSendCRC();                                 //Send 2 bytes of CRC

			data_response = ReadMedia();                //Read response

			if((data_response & 0x0F) != DATA_ACCEPTED)
			{
				status = FALSE;
			}
			else
			{
				index = 0;				//using i as a timeout counter

				do						//Wait for write completion
				{
					data_response = ReadMedia();
					index++;
				}while((data_response == 0x00) && (index != 0));

				if(index == 0)                                  //if timeout first
					status = FALSE;
			}

			mSend8ClkCycles();
		}

		SD_CS = 1;

	} // Not writing to 0 sector

	return(status);
} //end SECTORwrite


/*****************************************************************
 Return the state of the WP Pin
 *****************************************************************/
BYTE WriteProtectState(void)
{
	return(!SD_WE);
}

void Delayms(BYTE milliseconds)
{
#ifdef USE_PIC18
	T1CON = 0xB0;

	//enable interrupt
	PIE1bits.TMR1IE = 1;

	do {
		// load timer
		TMR1H = high(0xFFFF - MILLISECDELAY);
		TMR1L = low(0xFFFF - MILLISECDELAY);

		// clear flag
		PIR1bits.TMR1IF = 0;

		// turn on timer
		T1CONbits.TMR1ON = 1;

		// wait for it to be done
		while(!PIR1bits.TMR1IF)
		{
			;
		}

		// turn off timer
		T1CONbits.TMR1ON = 0;

		// Making a little more accurate
		Nop();
		Nop();

		milliseconds--;
	}while(milliseconds > 0);

Error:
	// now turn everything off 
	T1CONbits.TMR1ON = 0;
	PIE1bits.TMR1IE = 0;
#else
	BYTE 	ms;
	DWORD	count;
	
	ms = milliseconds;
	while (ms--)
	{
		count = ((SYSTEM_CLOCK / 28) / 1000) - 5;

		while (count--)
		{
			;
		}
	}
	Nop();
	return;
#endif
}


/******************************************************************************
 * Function:        BYTE ReadMedia(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          BYTE    - One byte of data read in from SPI port
 *
 * Side Effects:    None
 *
 * Overview:        ReadMedia reads in one byte of data while sending out 0xFF
 *
 * Note:            Could not use ReadSPI because it initializes SSPBUF to
 *                  0x00. The card expects 0xFF (see spec).
 *****************************************************************************/
BYTE ReadMedia(void)
{
	SPIBUF = 0xFF;                              //Data Out - Logic ones
	while(!SPISTAT_RBF);                     //Wait until cycle complete
	return(SPIBUF);                             //Return with byte read 
}//end ReadMedia


/******************************************************************************
 * Function:        BYTE StartupCard(DISK *dsk, CARDSTATUS *CardStatus)
 *
 * PreCondition:    SocketInitialize called
 *
 * Input:			dsk			- Disk to be mounted
 *
 * Output:			CE_GOOD				- Disk set up successfully
 *					CE_NOT_PRESENT		- Card not inserted
 *
 * Side Effects:    None
 *
 * Overview:        Startup the SDMMC Card
 *
 * Note:            None
 *****************************************************************************/

BYTE StartupCard(DISK *dsk, CARDSTATUS *CardStatus)
{
	CETYPE	        error;
	MMCSTATE        Flag;
	BYTE *	buffer = gblSDMMCBuf;
	CardStatus->_word = 0x00;

	CardStatus->CS_Reset      = TRUE;

	if((error = DISKmount( dsk, buffer, &Flag)) == CE_GOOD)
	{
		CardStatus->CS_Inserted   = TRUE;

		//set the type and write protect stuff
		CardStatus->CS_SDMMC  = ((Flag.isSDMMC) ? TRUE : FALSE);
		CardStatus->CS_WP     = ((Flag.isWP) ? TRUE : FALSE);
	}
	else
	{
		// there was an issue so turn off the card
		StopCard(dsk, CardStatus);

		CardStatus->CS_Failure = TRUE;
	}

	return(error);
}


/******************************************************************************
 * Function:        BYTE DISKmount ( DISK *dsk, BYTE* b, MMCSTATE *Flag)
 *
 * PreCondition:    SocketInitialize function has been executed.
 *
 * Input:           dsk		- The disk to be mounted
 *					b		- The buffer of to be assigned to the disk
 *					Flag	- Flag structure of card conditions
 *
 * Output:          CE_GOOD 		- Disk mounted
 *					CE_INIT_ERROR	- Initialization error has occured
 *
 * Side Effects:    None
 *
 * Overview:        Will mount only the first partition on the disk/card
 *
 * Note:            None
 *****************************************************************************/

BYTE DISKmount( DISK *dsk, BYTE* b, MMCSTATE *Flag)
{
//	BYTE          i;
	BYTE        error = CE_GOOD;

	dsk->mount = FALSE; // default invalid
	dsk->buffer = b;    // assign buffer

	// Initialize the sdmmc card
	if(MediaInitialize() != TRUE)
	{
		error = CE_INIT_ERROR;
	}
	else
	{
		// Load the Master Boot Record (partition)
		if((error = LoadMBR(dsk)) == CE_GOOD)
		{
			// Now the boot sector
			if((error = LoadBootSector(dsk)) == CE_GOOD)
				dsk->mount = TRUE; // Mark that the DISK mounted successfully
		}

	} // -- Initialize the SDMMC card

	return(error);
} // -- mount

/******************************************************************************
 * Function:        CETYPE LoadMBR ( DISK *dsk)
 *
 * PreCondition:    Called from DISKmount
 *
 * Input:           dsk		- The disk containing the master boot record to be loaded
 *
 * Output:          CE_GOOD 			- MBR loaded successfully
 *                  CE_BAD_SECTOR_READ	- A bad read occured of a sector 
 *					CE_BAD_PARTITION	- The boot record is bad
 *
 * Side Effects:    None
 *
 * Overview:        Loads the MBR and extracts the necessary information
 *
 * Note:            None
 *****************************************************************************/

BYTE LoadMBR(DISK *dsk)
{
    PT_FAT  Partition;
    BYTE error = CE_GOOD;
    BYTE type;

    // Get the partition table from the MBR
    if ( SECTORread( FO_MBR, dsk->buffer) != TRUE) 
        error = CE_BAD_SECTOR_READ;
    else
    {
		// assign it the partition table strucutre
    	Partition = (PT_FAT)dsk->buffer;
    
        // Ensure its good
        if((Partition->Signature0 != FAT_GOOD_SIGN_0) || (Partition->Signature1 != FAT_GOOD_SIGN_1))
        {
            error = CE_BAD_PARTITION;
        }
        else
        {    
            /*    Valid Master Boot Record Loaded   */

            // Get the 32 bit offset to the first partition 
            dsk->firsts = Partition->Partition0.PTE_FrstSect; 

            // check if the partition type is acceptable
           	type = Partition->Partition0.PTE_FSDesc;

            switch (type)
            {
                case 0x01:
                    dsk->type = FAT12;
                   	error = CE_CARDFAT12;
                    
                    break;
               
				case 0x04:
                case 0x06:
                case 0x0E:
                    dsk->type = FAT16;
                    break;
                
                case 0x0B:
                case 0x0C:
                    dsk->type = FAT32; // and error out				
                    error = CE_CARDFAT32;
                    break;
                    
                default:
                    error = CE_BAD_PARTITION;
            } // switch
        }
    }

    return(error);
}// -- LoadMBR


/******************************************************************************
 * Function:        BYTE LoadBootSector (DISK *dsk)
 *
 * PreCondition:    Called from DISKmount
 *
 * Input:           dsk		- The disk containing the boot sector
 *                  
 * Output:          CE_GOOD 			- Boot sector loaded
 *                  CE_BAD_SECTOR_READ	- A bad read occured of a sector 
 *					CE_NOT_FORMATTED	- The disk is of an unsupported format
 *					CE_CARDFAT12		- FAT12 during intial testing we are not supporting FAT12
 *					CE_CARDFAT32		- FAT 32 - card not supported
 *
 * Side Effects:    None
 *
 * Overview:        Load the boot sector information and extract the necessary information
 *
 * Note:            None
 *****************************************************************************/


BYTE LoadBootSector(DISK *dsk)
{
    WORD     RootDirSectors;
    DWORD     TotSec,DataSec;
    BYTE      error = CE_GOOD;
    BS		BootSec;        // boot sector, assume its FAT16 til we know better

	WORD   BytesPerSec;
                    
    // Get the Boot sector
    if ( SECTORread( dsk->firsts, dsk->buffer) != TRUE) 

        error = CE_BAD_SECTOR_READ;  
    else
    {
        // Assign the type
        BootSec = (BS)dsk->buffer;
    
        //Verify the Boot Sector is valid
        if((BootSec->Signature0 != FAT_GOOD_SIGN_0) || (BootSec->Signature1 != FAT_GOOD_SIGN_1))
        {
            error = CE_NOT_FORMATTED;
        }
        else
        {     
            // determine the number of sectors in one FAT (FATSz)
			#ifdef USE_PIC18
		        if(BootSec->FAT.FAT_16.BPB_FATSz16 != 0)
		        {
		            dsk->fatsize = BootSec->FAT.FAT_16.BPB_FATSz16;  
		        }
	            else
	            {
	                dsk->fatsize = BootSec->FAT.FAT_32.BPB_FATSz32;
	            }
			#else
				dsk->fatsize = ReadWord( dsk->buffer, BSI_FATSz16 ); 
			#endif

			// Figure out the total number of sectors
			#ifdef USE_PIC18
		        if(BootSec->FAT.FAT_16.BPB_TotSec16 != 0)
		        {
		            TotSec = BootSec->FAT.FAT_16.BPB_TotSec16;
		        }
	            else
	            {
	                TotSec = BootSec->FAT.FAT_16.BPB_TotSec32;
	            }
			#else
		        TotSec = ReadWord( dsk->buffer, BSI_TotSec16 );
				if( TotSec == 0 )
					TotSec = ReadDWord( dsk->buffer, BSI_TotSec32 );
			#endif
	            	                       

			#ifdef USE_PIC18
		        // 6.  get the full partition/drive layout
		        // 6.1 determine the size of a cluster
		        dsk->SecPerClus = BootSec->FAT.FAT_16.BPB_SecPerClus;
		        	        
		        // 6.2 determine fat, root and data lbas
		        // FAT = first sector in partition (boot record) + reserved records
		        dsk->fat        = dsk->firsts + BootSec->FAT.FAT_16.BPB_RsvdSecCnt;
		        
		        // 6.3 fatcopy is the number of FAT tables 
		        dsk->fatcopy    = BootSec->FAT.FAT_16.BPB_NumFATs;
	
		        // 6.4 MAX ROOT is the maximum number of entries in the root directory
		        dsk->maxroot    = BootSec->FAT.FAT_16.BPB_RootEntCnt;    
		        
	            // "root" Is the sector location of the root directory
	            dsk->root = dsk->fat + (dsk->fatcopy * dsk->fatsize);
	
				BytesPerSec = BootSec->FAT.FAT_16.BPB_BytsPerSec;
	
				if( BytesPerSec == 0 || (BytesPerSec & 1) == 1 ) //cannot be 0 or odd
					return( CE_NOT_FORMATTED );
		                        
		        RootDirSectors = ((BootSec->FAT.FAT_16.BPB_RootEntCnt * 32) + (BootSec->FAT.FAT_16.BPB_BytsPerSec - 1)) / BootSec->FAT.FAT_16.BPB_BytsPerSec;                    

			#else

				dsk->SecPerClus = ReadByte( dsk->buffer, BSI_SecPerClus );
		        	        
		        // 6.2 determine fat, root and data lbas
		        // FAT = first sector in partition (boot record) + reserved records
		        dsk->fat        = dsk->firsts + ReadWord( dsk->buffer, BSI_RsvdSecCount );
		        
		        // 6.3 fatcopy is the number of FAT tables 
		        dsk->fatcopy    = ReadByte( dsk->buffer, BSI_NumFATs );
	
		        // 6.4 MAX ROOT is the maximum number of entries in the root directory
		        dsk->maxroot    = ReadWord( dsk->buffer, BSI_RootEntCnt );   
		        
	            // "root" Is the sector location of the root directory
	            dsk->root = dsk->fat + (dsk->fatcopy * dsk->fatsize);
	
				BytesPerSec = ReadWord( dsk->buffer, BSI_BytesPerSec );
	
				if( BytesPerSec == 0 || (BytesPerSec & 1) == 1 ) //cannot be 0 or odd
					return( CE_NOT_FORMATTED );
		                        
		        RootDirSectors = ((dsk->maxroot * 32) + (BytesPerSec - 1)) / BytesPerSec;               
			#endif
	
			// figure out how many data sectors there is 
	        DataSec = TotSec - (dsk->root + RootDirSectors);
		        
		    dsk->maxcls = DataSec / dsk->SecPerClus;

	        // Straight out of MS FAT Hardware White Paper
	        if(dsk->maxcls < 4085)
	        {
	            /* Volume is FAT12 */
	            dsk->type = FAT12;
                error = CE_CARDFAT12;
    	    }
    	    else 
			{
				if(dsk->maxcls < 65525)
	    	    {
	    	        /* Volume is FAT16 */
	    	        dsk->type = FAT16;
	    	    }
	    	    else
	    	        /* Volume is FAT32 */
	    	        // We don't support FAT32
	    	        error = CE_CARDFAT32;     
			}
	                	        
	        // 6.5 DATA = ROOT + (MAXIMUM ROOT *32 / 512) 
	        dsk->data = dsk->root + ( dsk->maxroot >> 4); // assuming maxroot % 16 == 0!!!
	        
	        // make sure that we can read in a complete sector!
			#ifdef USE_PIC18
		        if(BootSec->FAT.FAT_16.BPB_BytsPerSec != MEDIA_SECTOR_SIZE)
			#else
		        if(BytesPerSec != MEDIA_SECTOR_SIZE)
			#endif
	            error = CE_NOT_FORMATTED;
		}
    }

	return(error);
}


/******************************************************************************
 * Function:        void StopCard(DISK *dsk, CARDSTATUS *cs)
 *
 * PreCondition:    Disk mounted
 *
 * Input:           cs		- Status of the card
 *					dsk		- Disk to be unmounted
 *                  
 * Output:			void
 *
 * Side Effects:    None
 *
 * Overview:        Cleanly unmount the card and fix the status
 *
 * Note:            None
 *****************************************************************************/

void StopCard(DISK *dsk, CARDSTATUS *cs)
{
    // turn off the card
    ShutdownMedia();
    
    cs->_word = 0x00;
    
    dsk->mount = FALSE;     
}


#ifdef USE_PIC24
	void CloseSPI( void )
	{
		SPI1STAT &= 0x7FFF;               // disable synchronous serial port
	}
	
	
	
	void OpenSPI( unsigned int sync_mode, unsigned char bus_mode, unsigned char smp_phase)
	{
	  SPISTAT = 0x0000;               // power on state 
	  SPICON = 0x0000;                // power on state
	  SPICON |= sync_mode;          // select serial mode 
	  SPICON |= smp_phase;          // select data input sample phase
	
	  switch( bus_mode )
	  {
	    case 0:                       // SPI1 bus mode 0,0
		#ifdef USE_PIC30
			SPI1CONbits.CKE = 1;
		#else
	      SPI1CON1bits.CKE = 1;       // data transmitted on rising edge
		#endif
	      break;    
	    case 2:                       // SPI1 bus mode 1,0
		#ifdef USE_PIC30
	      SPI1CONbits.CKP = 1;       // clock idle state high
		#else
	      SPI1CON1bits.CKP = 1;       // clock idle state high
		#endif
	      break;
	    case 3:                       // SPI1 bus mode 1,1
		#ifdef USE_PIC30
	      SPI1CONbits.CKP = 1;       // clock idle state high
		  SPI1CONbits.CKE = 0;
		#else
	      SPI1CON1bits.CKP = 1;       // clock idle state high
		  SPI1CON1bits.CKE = 0;
		#endif
	      break;
	    default:                      // default SPI1 bus mode 0,1
	      break;
	  }
	
	  switch( sync_mode )
	  {
	    case 4:                       // slave mode w /SS1 enable
	      TRISBbits.TRISB13 = 1;       // define /SS1 pin as input
	    case 5:                       // slave mode w/o /SS1 enable
	      TRISBbits.TRISB15 = 1;       // define clock pin as input
		#ifdef USE_PIC30
	      SPI1CONbits.SMP = 0;       // must be cleared in slave SPI mode
		#else
	      SPI1CON1bits.SMP = 0;       // must be cleared in slave SPI mode
		#endif
	      break;
	    default:                      // master mode, define clock pin as output
	      TRISBbits.TRISB15 = 0;       // define clock pin as output
	      break;
	  }
	
	  TRISB &= 0xBFFF;                  // define SDO1 as output (master or slave)
	  TRISB |= 0x1000;                  // define SDI1 as input (master or slave)
	  SPI1STATbits.SPIEN = 1;             // enable synchronous serial port 
	}
	
	
	unsigned char WriteSPI( unsigned char data_out )
	{
		BYTE	clear;
		SPIBUF = data_out;          // write byte to SSP1BUF register
		while( !SPI1STATbits.SPIRBF ); // wait until bus cycle complete
		clear = SPIBUF; 
		return ( 0 );                // return non-negative#
	}
	
	BYTE ReadByte( BYTE* pBuffer, WORD index )
	{
		return( pBuffer[index] );
	}
	
	WORD ReadWord( BYTE* pBuffer, WORD index )
	{
		BYTE loByte, hiByte;
		WORD res;
	
		loByte = pBuffer[index];
		hiByte = pBuffer[index+1];
		res = hiByte;
		res *= 0x100;
		res |= loByte;
		return( res );
	}
	
	
	DWORD ReadDWord( BYTE* pBuffer, WORD index )
	{
		WORD loWord, hiWord;
		DWORD result;
	
		loWord = ReadWord( pBuffer, index );
		hiWord = ReadWord( pBuffer, index+2 );
	
		result = hiWord;
		result *= 0x10000;
		result |= loWord;
		return result;
	}
#endif
